//
//    NSString+Autodoc.m 
//      Written by Carl Lindberg Copyright 1999 by Carl Lindberg.
//                       All rights reserved.
//        This notice may not be removed from this source code.
//
//	This header is included in the MiscKit by permission from the author
//	and its use is governed by the MiscKit license, found in the file
//	"License.rtf" in the MiscKit distribution.  Please refer to that file
//	for a list of all applicable permissions and restrictions.
//	
//  This file formatted with 4 spaces per tabstop.

#import "NSString+Autodoc.h"
#import "ADUtilities.h"

@implementation NSString (Autodoc)
/*"
 * This category exposes several of the Autodoc formatting functions as
 * NSString methods, so they can be easily called in specific circumstances
 * by templates as needed.
"*/

/*"
 * Passes the receiving string through ADFormatString(), which will
 * reformat any Autodoc markup in the string, bold certain words like "id"
 * and "self", and do some output-format-specific formatting.  This is the
 * function that comments are normally passed through.
"*/
- (NSString *)formattedMarkupString
{
    return ADFormatString(self, nil);
}

/*"
 * Passes the receiving string through ADFormatValueString(), which indents
 * the string according to how many '{' and '}' characters exist and does
 * output-format-specific formatting.  This is the function that variable
 * "type" strings are normally passed through.
"*/
- (NSString *)formattedString
{
    return ADFormatValueString(self);
}

/*"
 * Prefixes each line in the receiving string with the outputModule's
 * -#valueIndentString.
"*/
- (NSString *)indentedString
{
    return ADIndentString(self, 1, YES);
}

/*"
 * Like -#indentedString, except the first line is NOT indented.  Useful if
 * a template needs to indent a string, but already does special indenting
 * on the first line.
"*/
- (NSString *)indentedStringExceptFirstLine
{
    return ADIndentString(self, 1, NO);
}

@end
