#!/usr/bin/perl
#
# formatComment.pl
# 
# Copyright (c) 1999 Rhyme & Reason, LLC. All rights reserved
# No warrantees of any kind, expressed or implied. You are free use distribute
# modify and reuse (though you may not want to) this code. 
#
# Author: K. Lo Shih, lo@rhymereason.com
#
#
# Here's a recent rewrite of format_comment that I originally wrote back in the
# days of Adam and bbum at Apollo where all autodoc, I lost track of my original
# one. It's much better and more configurable and as good as I can do given my
# patience with Perl and without making a full-blown perlobj mess and tons of
# package dependencies.
#
# The major difference between this version and previous ones is that it's
# relatively configurable to customize existing comment styles and create new
# ones. See the section on CONFIGURATION below.
# 
# I haven't really used bbum's version so I'm not really I familiar with it. but
# it's at MiscKit (of.org) for you reference, named format_comment sitting next 
# to autodoc.
# 
# (K. Lo Shih, lo@rhymereason.com)

    $program = ($0 =~ /([^\/]+)$/)[0]; # Cheap basename(), get use to it. The
                                       # whole script is going to be cheap.
    $copyright = "1999 Rhyme & Reason";
    $module_version = '$Id: formatComment.pl,v 1.6 1999/11/06 13:04:27 kls Exp $';
    $help_usage = <<END;
Usage: $program [<options>]

Options:
  -d --debug        Show debugging information
  -t --test         Test (currently incomplete)
  -v --version      Show version
  -h --help         Show this help page
END

    $isDebugging = 0;
    $" = ", ";
    $maximumLineLength = 80;
    $tabWidth = 4;

    # Parse arguments 
    while ($_ = shift) {
        /^(-d|--debug)$/ && do { $isDebugging = 1; next; };
        /^(-t|--test)$/ && do { test(); exit; };
        /^(-v|--version)$/i && do { version(); exit; };
        /^(-h|--help)$/i && do { help(); exit; };
        /^-/ && do { print STDERR "Unrecognized switch \"$_\"\n"; $argerror++; 
            next; };
        do { print "Unrecognized argument \"$_\"\n"; $argerror++; next; };
    }
    help() if $argerror;


    # 
    # CONFIGURATION:
    #
    # The following stanzas are used to configure different kinds of comments
    # that should be handle. Modify the as you will, add support in the engine
    # below to support new features. Currently, this script supports:
    #
    # 'name' (string)
    # This is the name of the comment type for debugging purposes.
    #
    # 'detectRegex' (regex)
    # This regex is used for detecting if this comment type is requested. Your
    # only requirement is that you must use one and only one () grouping 
    # parenthesis. (This is because of the way that this script matches the
    # regex to the comment type. We can fix this later if necessary.) The 
    # detect regex for all comment types are joined together with a '|' OR
    # character when the script is trying to detect which comment type to use.
    # The script searches for the comment types in the order they appear in
    # the @commentTypes array.
    #
    # 'getLeadingWhiteSpaceRegex' (regex)
    # This regular expression find how much white space there is at the 
    # beginning of the detect line and uses that calculation for the rest of
    # the script for outputting lines.
    #
    # 'substitutionRegex' (regex)
    # This regex is used for substituting variable length or character strings
    # from the initial detect line for some of the prefix values. For example,
    # 'CondensedCStyleCommentType' uses this feature to accept comments with
    # any number of starts /****** as a prefix and appends that prefix to the
    # beginning of every output line.
    #
    # 'getLineTextRegex' (regex)
    # This regex is what you use to extract the text content of the comment. It
    # should strip off leading comment structure characters like /* or *. It
    # should always have one grouping () that is used to return the text of the
    # comment line. Make sure you get this right so that you can run this 
    # script on a comment over and over without it changing.
    #
    # 'headerLinePrefix' (string, regex sub)
    # 'bodyLinePrefix' (string, regex sub)
    # 'footerLinePrefix' (string, regex sub)
    # These strings are the prefix characters without leading white space for
    # headers. This prefix is the first thing on the line. See the you can put
    # in variable substitutions in these strings. They're eval()ed before
    # they're used. In particular, you can use the $1 grouping matches from the
    # 'substitutionRegex' which is performed just before they're eval()ed.
    #
    # 'hasHeader' (boolean)
    # 'hasFooter' (boolean)
    # These are used to determine if there are different header/footer markers
    # for the comment. // and # comments don't have headers or footers. /* */
    # style comments do.
    #
    # 'condensedHeader' (boolean)
    # 'condensedFooter' (boolean)
    # 'footerLineSuffixOnNewLine' (boolean)
    # The 'condensed...' values are true if the header or footer shouldn't be on
    # a line by itself. Rather, the condensed header uses the 'headerLinePrefix'
    # in place of the 'bodyLinePrefix' on the first line of comment text. The
    # condensed footer uses a 'bodyLinePrefix' on the last line of comment text
    # but appends a 'footerLinePrefix' at the very end of the comment. (Sorry
    # for the misnamed var here.)
    #
    # 'headerLineSuffix' (string)
    # 'bodyLineSuffix' (string)
    # 'footerLineSuffix' (string)
    # 'headerBarCharacter' (one-character string)
    # 'footerBarCharacter' (one-character string)
    # The suffixes if they exist allow you to append a string at the right edge
    # of a comment. If there are headers or footers and there is a bar character
    # specified, this also constructs a bar across the top and/or bottom. You
    # can form a box around a comment with this mechanism.
    #
    # 'bodyTextJustification' (string)
    # Justification is only performed if there is a 'bodyLineSuffix' defined.
    # otherwise the text could be too far away from the comment marker that it
    # just doesn't look good. Of course, if someone wants this, I guess we
    # could put it in.

    @commentTypes = (
        {
            #
            # /* Hello      -->    /* Hello */
            #
            # /*** Hello    -->    /*** Hello */
            #
            'name' => 'CondensedCStyleCommentType',
            'detectRegex' => '(/\*+)(?= \w)',
            'substitutionRegex' => '/(\*+)',
            'headerLinePrefix' => '/$1',
            'bodyLinePrefix' => ' $1 ',
            'footerLinePrefix' => '*/', # Add a space if you don't condense
            'getLeadingWhiteSpaceRegex' => '^(\s+)',
            'getLineTextRegex' => '^[\s/*]*(.*?)(\*/)?\s*$',
            'hasHeader' => 1,
            'condensedHeader' => 1,
            'hasFooter' => 1,
            'condensedFooter' => 1,  # Change this to '0' if you like hanging
                                     # footers
            'bodyTextJustification' => 'left',
        },
        {
            #
            # /***           -->     /***
            # Hello                   *** Hello
            #                         ***/
            'name' => 'VariableAsteriskStandardCommentType',
            'detectRegex' => '(/\*{1,10})(?![*"=-]| =|>-|\+-)',
            'substitutionRegex' => '/(\*+)',
            'headerLinePrefix' => '/$1 ',
            'bodyLinePrefix' => ' $1 ',
            'footerLinePrefix' => ' $1/',
            'getLeadingWhiteSpaceRegex' => '^(\s+)',
            'getLineTextRegex' => '^[\s/*]*(.*?)(\*/)?\s*$',
            'hasHeader' => 1,
            'hasFooter' => 1,
            'bodyTextJustification' => 'left',
        },
        {
            #                        /*"
            # /*" Hello          -->  * Hello
            #                        "*/
            'name' => 'AutodocStandardCommentType',
            'detectRegex' => '(/\*"|"")',
            'getLeadingWhiteSpaceRegex' => '^(\s+)',
            'getLineTextRegex' => '^\s*(?:/\*"|"\*/|\*/|\*|"")?\s*(.*)$',
            'hasHeader' => 1,
            'hasFooter' => 1,
            'bodyLinePrefix' => ' * ',
            'headerLinePrefix' => '/*"',
            'footerLinePrefix' => '"*/',
            'bodyTextJustification' => 'left',
        },
        {
            # (more than 10 asterisks, triggers it)
            #
            #                                /**************************
            # /************* Hello       -->  * Hello                  *
            #                                 **************************/
            'name' => 'AsteriskBoxCommentType',
            'detectRegex' => '(/\*{11,})',
            'getLeadingWhiteSpaceRegex' => '^(\s+)',
            'getLineTextRegex' => '^[\s/*]*(.*?)(\*/|\*)?\s*$',
            'hasHeader' => 1,
            'hasFooter' => 1,
            'headerLinePrefix' => '/*',
            'headerLineSuffix' => '* ',
            'bodyLinePrefix' => ' * ',
            'bodyLineSuffix' => '* ',
            'footerLinePrefix' => ' *',
            'footerLineSuffix' => '*/',
            'headerBarCharacter' => '*',
            'footerBarCharacter' => '*',
            'bodyTextJustification' => 'left',
        },
        {
            #
            #                                /*------------------------*
            # /*- Hello                  -->  | Hello                  |
            #                                 *------------------------*/
            'name' => 'LineBoxCommentType',
            'detectRegex' => '(/\*-)',
            'getLeadingWhiteSpaceRegex' => '^(\s+)',
            'getLineTextRegex' => '^[\s/*|-]*(.*?)(\*/|\*|\|)?\s*$',
            'hasHeader' => 1,
            'hasFooter' => 1,
            'headerLinePrefix' => '/*',
            'headerLineSuffix' => '* ',
            'bodyLinePrefix' => ' | ',
            'bodyLineSuffix' => ' | ',
            'footerLinePrefix' => ' *',
            'footerLineSuffix' => '*/',
            'headerBarCharacter' => '-',
            'footerBarCharacter' => '-',
            'bodyTextJustification' => 'left',
        },
        {
            #                                /*>-----------------------*
            # /*>- Hello                 -->  |                  Hello |
            #                                 *------------------------*/
            'name' => 'RightJustifiedLineBoxCommentType',
            'superclass' => 'LineBoxCommentType',
            'detectRegex' => '(/\*>-)',
            'headerLinePrefix' => '/*>',
            'getLineTextRegex' => '^[\s/*|>-]*(.*?)(\*/|\*|\|)?\s*$',
            'bodyTextJustification' => 'right',
        },
        {
            #                                /*+-----------------------*
            # /*+- Hello                 -->  |         Hello          |
            #                                 *------------------------*/
            'name' => 'CenterJustifiedLineBoxCommentType',
            'superclass' => 'LineBoxCommentType',
            'detectRegex' => '(/\*\+-)',
            'headerLinePrefix' => '/*+',
            'getLineTextRegex' => '^[\s/*|+-]*(.*?)(\*/|\*|\|)?\s*$',
            'bodyTextJustification' => 'center',
        },
        {
            #                               /**=========================
            # /**= Hello                -->  ** Hello
            #                                **=========================
            #                                **/
            'name' => 'EqualSignBarWithFooterCommentType',
            'detectRegex' => '(/\*+=)',
            'substitutionRegex' => '/(\*+)',
            'headerLinePrefix' => '/$1',
            'bodyLinePrefix' => ' $1 ',
            'footerLinePrefix' => ' $1',
            'getLeadingWhiteSpaceRegex' => '^(\s+)',
            'getLineTextRegex' => '^[\s/*=]*(.*?)(\*/|\*)?\s*$',
            'hasHeader' => 1,
            'hasFooter' => 1,
            'headerLineSuffix' => '',
            'footerLineSuffix' => ' $1/',
            'footerLineSuffixOnNewLine' => 1,
            'headerBarCharacter' => '=',
            'footerBarCharacter' => '=',
            'bodyTextJustification' => 'left',
        },
        {
            #                               /*** ========================
            # /*** = Hello              --> **** Hello
            #                               **** ========================
            #                               ***/
            'name' => 'UglyEqualSignBarWithFooterCommentType',
            'detectRegex' => '(/\*+ =)',
            'substitutionRegex' => '/(\*+)',
            'headerLinePrefix' => '/$1 ',
            'bodyLinePrefix' => '*$1 ',
            'footerLinePrefix' => '*$1 ',
            'getLeadingWhiteSpaceRegex' => '^(\s+)',
            'getLineTextRegex' => '^[\s/*=]*(.*?)(\*/|\*)?\s*$',
            'hasHeader' => 1,
            'hasFooter' => 1,
            'headerLineSuffix' => '',
            'footerLineSuffix' => '$1/',
            'footerLineSuffixOnNewLine' => 1,
            'headerBarCharacter' => '=',
            'footerBarCharacter' => '=',
            'bodyTextJustification' => 'left',
        },
        {
            #
            # //  Hello. Here's              // Hello. Here's some text
            # some text that wraps       --> // that wraps a couple of
            # a couple of lines.             // lines.
            # 
            'name' => 'DoubleSlashCommentType',
            'detectRegex' => '(//)',
            'getLeadingWhiteSpaceRegex' => '^(\s+)',
            'getLineTextRegex' => '^[\s//*]*(.*)$',
            'hasHeader' => 0,
            'hasFooter' => 0,
            'bodyLinePrefix' => '// ',
            'headerLinePrefix' => '',
            'footerLinePrefix' => '',
            'bodyTextJustification' => 'left',
        },
        {
            #
            # #  Hello. Here's               # Hello. Here's some text
            # some text that wraps       --> # that wraps a couple of
            # a couple of lines.             # lines.
            # 
            'name' => 'ScriptCommentType',
            'detectRegex' => '(#)',
            'getLeadingWhiteSpaceRegex' => '^(\s+)',
            'getLineTextRegex' => '^[\s#]*(.*)$',
            'hasHeader' => 0,
            'hasFooter' => 0,
            'bodyLinePrefix' => '# ',
            'headerLinePrefix' => '#',
            'footerLinePrefix' => '#',
            'bodyTextJustification' => 'left',
        },
        {
            #
            # /*             -->     /*
            # Hello                   * Hello
            #                         */
            'name' => 'StandardCommentType',
            'detectRegex' => '(/\*)(?!["\'])',
            'getLeadingWhiteSpaceRegex' => '^(\s+)',
            'getLineTextRegex' => '^[\s/*]*(.*?)(\*/)?\s*$',
            'hasHeader' => 1,
            'hasFooter' => 1,
            'bodyLinePrefix' => ' * ',
            'headerLinePrefix' => '/* ',
            'footerLinePrefix' => ' */',
            'bodyTextJustification' => 'left',
        },
    );
#    $DefaultCommentType = $commentTypes[2]; # Autodoc
    $DefaultCommentType = $commentTypes[11]; # Standard C

    # The %{$commentTypes} keys to perform an eval on 
    @prefixKeysToEval = (
        'headerLinePrefix',
        'headerLineSuffix',
        'bodyLinePrefix',
        'bodyLineSuffix',
        'footerLinePrefix',
        'footerLineSuffix',
    );

    ### Come up with a dictionary based on name.
    map { $commentTypes{$_->{'name'}} = $_; } @commentTypes;
    ### Perform subclassing. Inheritance is only one level deep. sorry.
    map {     
        if (defined($_->{'superclass'})) {
            $superclass = $commentTypes{$_->{'superclass'}};
            foreach $key (keys %{$superclass}) {
                $_->{$key} = $superclass->{$key} unless defined $_->{$key};
            }
        }
    } @commentTypes;

    #
    # Collect all of the comment type detection regexes from the comment type
    # configuration.
    #
    map { push(@detectRegexes, $_->{'detectRegex'}) unless $_->{'disabled'} }
        @commentTypes;
    $detectRegexes = join('|', @detectRegexes);

    
    #
    # Extract out all the lines from the input
    #
    @inputLines = <STDIN>;
    @outputLines = ();
    
    $commentTypeDetected = 0;
    
    #
    # Extract the following information from the @inputLines:
    #
    # $commentType - (hash reference) the type of the comment
    # $leadingWhiteSpace - (string) the space in front of each output line
    # @paragraphs - (array of references to arrays of strings) a collection
    #     of all of the paragraphs and the words in them.
    #
    @paragraphs = ();
    $currentParagraph = [];
    while ($_ = shift(@inputLines)) {
        chop();
        
        #
        # Always detect the comment first.
        #
        if (! $commentTypeDetected) {
            #
            # Attempt to detect the type of comment it is. Use all of the
            # comment types' 'detectRegex' expressions (collected in 
            # $detectRegexes. If we cannot find a match, use the 
            # $DefaultCommentType.
            #
            debug("Detecting comment type using regex: %s", $detectRegexes);
            debug("Detecting using input line: \"$_\"");
            if (@detectedRegexMatches = m|$detectRegexes|) {

                debug("Detected matches: @detectedRegexMatches");
                $index = 0;
                foreach $match (@detectedRegexMatches) {
                    debug("Checking type[%d]: %s", $index,
                        $commentTypes[$index]->{'name'});
                    if (length($match) && !$commentTypeDetected) {
                        # 
                        # Found a match. First evaluate the prefix definitions
                        # just in case some of them contain references to the
                        # detect regex. So we have to a m|| on the regex 
                        # again.
                        # 
                        $commentType = $commentTypes[$index];
                        debug("Detected type: %s", $commentType->{'name'});
                        m|$commentType->{'substitutionRegex'}|;
                        @definedEvalPrefixKeys = grep(defined(
                            $commentType->{$_}), @prefixKeysToEval);
print @evalPrefixKeys;
                        map { 
                            $_ = eval(qq(qq($_))); 
                        } @{$commentType}{@definedEvalPrefixKeys};
                        $commentTypeDetected = 1;
                        break;
                    }
                    $index++;
                    break if $commentTypeDetected;
                }
            }
            if (! $commentTypeDetected) {
                $commentTypeDetected = 1;
                $commentType = $DefaultCommentType;
                debug("Using default comment type: %s", 
                    $commentType->{'name'});
            }
            
            
            #
            # Get the leading white space from the first line. We'll use it
            # to prefix all of our output lines.
            #
            $getLeadingWhiteSpaceRegex = 
                $commentType->{'getLeadingWhiteSpaceRegex'};
            ($leadingWhiteSpace) = m|$getLeadingWhiteSpaceRegex|;
            $leadingWhiteSpace =~ s/\t/    /g;
        } 

        #
        # Collect the comment text using the comment type's regular expression
        #
        $getLineTextRegex = $commentType->{'getLineTextRegex'};
        ($lineText) = m|$getLineTextRegex|;
        debug("Extracted line text: \"%s\"", $lineText);
        
        #
        # If the line text is not empty, then add the words to the current
        # paragraph. Go on to a new paragraph.
        #
        if ($lineText !~ /\S/) {
            #
            # Add a blank paragraph if this is not the first paragraph and
            # the last paragraph was not empty (first word is not empty).
            #
            push(@paragraphs, $currentParagraph);
            push(@paragraphs, ['']) if (1+$#paragraphs > 0)
                    && length(${$paragraphs[$#paragraphs]}[0]) > 0;
            $currentParagraph = [];
            debug("Detected new paragraph");
        } else {
            push(@{$currentParagraph}, grep(/\S/, split(/\s+/, $lineText)));
            debug("Detected new words, total in this paragraph: %d", 
                $#{$currentParagraph} + 1);
        }
    }
    
    #
    # If we leave any words in the current paragraph roll them into our
    # list of paragraphs.
    #
    if (1+$#{$currentParagraph} > 0) {
        push(@paragraphs, $currentParagraph);
    }
    
    #
    # If the last paragraph is empty, trim it off.
    #
    while (1+$#paragraphs > 0 && length(${$paragraphs[$#paragraphs]}[0]) == 0) {
        debug("Trimming off last empty paragraph");
        pop(@paragraphs);
    }    
    
    #
    # If the comment type requires a header, add it to the output lines.
    # 
    if ($commentType->{'hasHeader'}) {
        if ($commentType->{'condensedHeader'}) {
            $currentOutputLine = sprintf("%s%s", $leadingWhiteSpace, 
                $commentType->{'headerLinePrefix'});
        } 
        #
        # Support for horizontal bars on top of comment 
        #
        elsif (defined $commentType->{'headerLineSuffix'}) {
            ### Create the line like usual
            $currentOutputLine = sprintf("%s%s", $leadingWhiteSpace, 
                $commentType->{'headerLinePrefix'});
            ### ...but add a horizontal bar just long enough.
            $barCharacter = $commentType->{'headerBarCharacter'} || ' ';
            $horizontalBar = $barCharacter x ($maximumLineLength - 
                length($currentOutputLine) - 
                length($commentType->{'headerLineSuffix'}));
            ### ...to fit the headerLineSuffix, too.
            $currentOutputLine .= ($horizontalBar . 
                $commentType->{'headerLineSuffix'});
            push(@outputLines, $currentOutputLine);
            $currentOutputLine = "";
        } 
        #
        # Standard non-condensed or non-header comments
        #
        else {
            push(@outputLines, sprintf("%s%s", $leadingWhiteSpace, 
                $commentType->{'headerLinePrefix'}));
        }
    }

    #
    # Handle condensed footers
    # 
    if ($commentType->{'hasFooter'} && $commentType->{'condensedFooter'}) {
        push(@{$paragraphs[$#paragraphs]}, $commentType->{'footerLinePrefix'});
    }
    
    # 
    # Lo, wouldn't it be cool if you would just make the output into an object
    # It would be so much easier. Of course, it'd be even better to do this in
    # something other than Perl, but who has the time to write this in Python?
    # Excuses aside, this code is a little tangly. Sorry about that.
    # 
    # Loops through all @paragraphs collected previously, looping through each
    # $word. Appending them to the $currentOutputLine. When the line gets too
    # long, the line is prepared and pushed onto @outputLines which is a buffer
    # for the final output. This loop only handles the body of the comment.
    # Headers and footers are handled above and below.
    #
    $maximumBodyLineLength = $maximumLineLength - 
        length($commentType->{'bodyLineSuffix'});
    debug("Body line suffix: \"%s\"", $commentType->{'bodyLineSuffix'});
    debug("maximumBodyLineLength: \"%s\"", $maximumBodyLineLength);
    foreach $paragraph (@paragraphs) {
        foreach $word (@{$paragraph}) {
            if (length($currentOutputLine) == 0) {
                $currentOutputLine = sprintf("%s%s%s", $leadingWhiteSpace,
                    $commentType->{'bodyLinePrefix'}, $word);
            } else {
                $proposedOutputLine = $currentOutputLine . " " . $word;
                debug("Proposed output line: \"%s\"", $proposedOutputLine);

                ### Test to see if the line is too long
                if (length($proposedOutputLine) > $maximumBodyLineLength) {
                    # End of the line inside our paragraph
                    ### Handle bodyLineSuffixes now.
                    if (defined $commentType->{'bodyLineSuffix'}) {
                        $remainingSpaceLength = $maximumBodyLineLength - 
                            length($currentOutputLine);
                        ### It's okay if $remainingSpaceLength < 0
                        $remainingSpace = ' ' x $remainingSpaceLength;
                        ### Handle justification
                        $bodyTextSplitIndex = length($leadingWhiteSpace) +
                            length($commentType->{'bodyLinePrefix'});
                        if ($commentType->{'bodyTextJustification'} eq
                            'right') {
                            $currentOutputLine = substr($currentOutputLine, 0,
                                $bodyTextSplitIndex) . $remainingSpace . 
                                substr($currentOutputLine, $bodyTextSplitIndex)
                                . $commentType->{'bodyLineSuffix'};
                        } 
                        elsif ($commentType->{'bodyTextJustification'} eq 
                            'center') {
                            $leftSpace = substr($remainingSpace, 0, length(
                                $remainingSpace) / 2);
                            $rightSpace = substr($remainingSpace, length(
                                $leftSpace));
                            $currentOutputLine = substr($currentOutputLine, 0,
                                $bodyTextSplitIndex) . $leftSpace . 
                                substr($currentOutputLine, $bodyTextSplitIndex)
                                . $rightSpace .$commentType->{'bodyLineSuffix'};
                        }
                        else {
                            $currentOutputLine .= $remainingSpace . 
                                $commentType->{'bodyLineSuffix'};
                        }
                    }

                    ### Write the line to our cached output...
                    push(@outputLines, $currentOutputLine);
                    $currentOutputLine = "";
                    redo;
                } else {
                    ### Line's not long enough, next word...
                    $currentOutputLine = $proposedOutputLine;
                }
            }
        }
        # Process any last lines in the paragraph...
        if (length($currentOutputLine)) {
            # (Sorry, have to do the same code again here... )
            # (Sorry, have to do the same code again here... )
            # (Sorry, have to do the same code again here... )
            # Because of the design of this foreach block, I had to write this
            # twice. I'll come back and clean this up. Either strip it out
            # into a function or unravel the context so that the last 
            # parapgraph line can be handled along with the others.
            # 
            # Handle bodyLineSuffixes now.
            if (defined $commentType->{'bodyLineSuffix'}) {
                $remainingSpaceLength = $maximumBodyLineLength -
                    length($currentOutputLine);
                ### It's okay if $remainingSpaceLength < 0
                $remainingSpace = ' ' x $remainingSpaceLength;
                ### Handle justification
                $bodyTextSplitIndex = length($leadingWhiteSpace) +
                    length($commentType->{'bodyLinePrefix'});
                if ($commentType->{'bodyTextJustification'} eq
                    'right') {
                    $currentOutputLine = substr($currentOutputLine, 0,
                        $bodyTextSplitIndex) . $remainingSpace . 
                        substr($currentOutputLine, $bodyTextSplitIndex)
                        . $commentType->{'bodyLineSuffix'};
                } 
                elsif ($commentType->{'bodyTextJustification'} eq 
                    'center') {
                    $leftSpace = substr($remainingSpace, 0, length(
                        $remainingSpace) / 2);
                    $rightSpace = substr($remainingSpace, length(
                        $leftSpace));
                    $currentOutputLine = substr($currentOutputLine, 0,
                        $bodyTextSplitIndex) . $leftSpace . 
                        substr($currentOutputLine, $bodyTextSplitIndex)
                        . $rightSpace .$commentType->{'bodyLineSuffix'};
                }
                else {
                    $currentOutputLine .= $remainingSpace . 
                        $commentType->{'bodyLineSuffix'};
                }
            }
            ### push the current line to the buffer. Go onto the next paragraph
             push(@outputLines, $currentOutputLine);
            $currentOutputLine = "";
        }
    }

    #
    # If the comment type requires a footer, add it to the output lines.
    #
    if ($commentType->{'hasFooter'} && !$commentType->{'condensedFooter'}) {
        ### Handle defined bars on the bottom of a comment
        if (defined $commentType->{'footerLineSuffix'}) {
            ### Create the line like usual
            $currentOutputLine = sprintf("%s%s", $leadingWhiteSpace,
                $commentType->{'footerLinePrefix'});
            ### ...but add a horizontal bar just long enough.
            $barCharacter = $commentType->{'footerBarCharacter'} || ' ';
            $barLength = $maximumLineLength - length($currentOutputLine);
            ### Make the line longer if the suffix is going on a new line.
            if (! $commentType->{'footerLineSuffixOnNewLine'}) {
                $barLength -= length($commentType->{'footerLineSuffix'});
            }
            $horizontalBar = $barCharacter x $barLength;
            ### ...to fit the headerLineSuffix, too.
            $currentOutputLine .= $horizontalBar;
            ### If the footer suffix is supposed to be on a new line...
            if ($commentType->{'footerLineSuffixOnNewLine'}) {
                push(@outputLines, $currentOutputLine);
                $currentOutputLine = $leadingWhiteSpace;
            }
            $currentOutputLine .= $commentType->{'footerLineSuffix'};
            push(@outputLines, $currentOutputLine);
        }
        ### All other kinds of comments
        else {
            push(@outputLines, sprintf("%s%s", $leadingWhiteSpace,
                $commentType->{'footerLinePrefix'}));
        }
    }
    
    print join("\n", @outputLines), "\n";
    

#
# Synopsys: debug("format", ...);
#
sub debug {
    return unless $isDebugging;
    printf STDOUT @_;
    printf STDOUT "\n";
}
    
# Incomplete
sub test {
    %testSuite = eval(<DATA>);
    die "eval: $@\n" if $@;
    foreach $test (keys %testSuite) {
        print $test;
    }
}

sub help {
    print $help_usage;
    exit 1;
}

sub version {
    print <<END;
$0

$module_version
No warranties. Copyright (c) $copyright
END
    exit;
}

